@@ -4,11 +4,14 @@
 use std::borrow::Cow;
 use std::sync::Arc;
 
+use http::HeaderValue;
+
 use crate::credential::{
-    AuthenticationCredentialToken, AuthenticationCredentialUsernamePassword, HttpRealmCredentials,
+    AuthenticationCredential, FetchedToken, FetchedUsernamePassword, HttpRealmCredentials,
+};
+use crate::{
+    AuthenticError, AuthenticationProtocol, AuthenticationProtocolConfigure, AuthenticationStep,
 };
-use crate::sensitive::SetSensitiveHeader;
-use crate::{AuthenticError, AuthenticationProtocol, AuthenticationStep};
 
 /// Protocol for no authentication
 ///
@@ -23,12 +26,13 @@ impl NoAuthentication {
 }
 
 impl AuthenticationProtocol for NoAuthentication {
-    type Builder = http::request::Builder;
     type Request = hyper::Request<hyper::Body>;
     type Response = hyper::Response<hyper::Body>;
     type Error = hyper::Error;
 }
 
+impl AuthenticationProtocolConfigure<http::request::Builder> for NoAuthentication {}
+
 /// Authentication using a token in a specified header.
 pub struct HeaderAuthentication<Credential> {
     header_name: Cow<'static, str>,
@@ -37,27 +41,48 @@ pub struct HeaderAuthentication<Credential> {
 
 impl<Credential: 'static> HeaderAuthentication<Credential>
 where
-    Credential: AuthenticationCredentialToken,
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedToken,
 {
-    pub fn new(header_name: impl Into<Cow<'static, str>>, credential: &Arc<Credential>) -> Self {
+    pub fn new(header_name: impl Into<Cow<'static, str>>, credential: Arc<Credential>) -> Self {
         Self {
             header_name: header_name.into(),
-            credential: credential.clone(),
+            credential,
         }
     }
 }
 
 impl<Credential> AuthenticationProtocol for HeaderAuthentication<Credential>
 where
-    Credential: AuthenticationCredentialToken,
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedToken,
 {
-    type Builder = http::request::Builder;
     type Request = hyper::Request<hyper::Body>;
     type Response = hyper::Response<hyper::Body>;
     type Error = hyper::Error;
 
-    fn configure(&self, builder: Self::Builder) -> Self::Builder {
-        builder.set_sensitive_header(self.header_name.as_ref(), self.credential.token())
+    fn step(&self) -> Result<Option<AuthenticationStep<Self::Request>>, AuthenticError> {
+        match self.credential.auth_step() {
+            Ok(duration) if duration.is_zero() => Ok(None),
+            Ok(duration) => Ok(Some(AuthenticationStep::WaitFor(duration))),
+            Err(err) => Err(err),
+        }
+    }
+}
+
+impl<Credential> AuthenticationProtocolConfigure<http::request::Builder>
+    for HeaderAuthentication<Credential>
+where
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedToken,
+{
+    fn configure(
+        &self,
+        builder: http::request::Builder,
+    ) -> Result<http::request::Builder, AuthenticError> {
+        let mut header_value = HeaderValue::try_from(self.credential.fetch()?.token())?;
+        header_value.set_sensitive(true);
+        Ok(builder.header(self.header_name.as_ref(), header_value))
     }
 }
 
@@ -67,14 +92,15 @@ pub struct BearerAuthentication<Credential> {
     credential: Arc<Credential>,
 }
 
-impl<Credential: 'static> BearerAuthentication<Credential>
+impl<Credential> BearerAuthentication<Credential>
 where
-    Credential: AuthenticationCredentialToken,
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedToken,
 {
-    pub fn new(credential: &Arc<Credential>) -> Self {
+    pub fn new(credential: Arc<Credential>) -> Self {
         Self {
             auth_scheme: "Bearer".into(),
-            credential: credential.clone(),
+            credential,
         }
     }
 
@@ -90,20 +116,41 @@ where
 
 impl<Credential> AuthenticationProtocol for BearerAuthentication<Credential>
 where
-    Credential: AuthenticationCredentialToken,
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedToken,
 {
-    type Builder = http::request::Builder;
     type Request = hyper::Request<hyper::Body>;
     type Response = hyper::Response<hyper::Body>;
     type Error = hyper::Error;
 
-    fn configure(&self, builder: Self::Builder) -> Self::Builder {
-        let token = self.credential.token();
+    fn step(&self) -> Result<Option<AuthenticationStep<Self::Request>>, AuthenticError> {
+        match self.credential.auth_step() {
+            Ok(duration) if duration.is_zero() => Ok(None),
+            Ok(duration) => Ok(Some(AuthenticationStep::WaitFor(duration))),
+            Err(err) => Err(err),
+        }
+    }
+}
+
+impl<Credential> AuthenticationProtocolConfigure<http::request::Builder>
+    for BearerAuthentication<Credential>
+where
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedToken,
+{
+    fn configure(
+        &self,
+        builder: http::request::Builder,
+    ) -> Result<http::request::Builder, AuthenticError> {
+        let fetched = self.credential.fetch()?;
+        let token = fetched.token();
         let mut value = Vec::with_capacity(self.auth_scheme.len() + 1 + token.len());
         value.extend(self.auth_scheme.as_bytes());
         value.push(b' ');
         value.extend(token);
-        builder.set_sensitive_header(hyper::header::AUTHORIZATION, &value[..])
+        let mut header_value = HeaderValue::try_from(value)?;
+        header_value.set_sensitive(true);
+        Ok(builder.header(hyper::header::AUTHORIZATION, header_value))
     }
 }
 
@@ -112,32 +159,45 @@ pub struct BasicAuthentication<Credential> {
     credential: Arc<Credential>,
 }
 
-impl<Credential> BasicAuthentication<Credential>
-where
-    Credential: AuthenticationCredentialUsernamePassword + 'static,
-{
-    pub fn new(credential: &Arc<Credential>) -> Self {
-        Self {
-            credential: credential.clone(),
-        }
+impl<Credential> BasicAuthentication<Credential> {
+    pub fn new(credential: Arc<Credential>) -> Self {
+        Self { credential }
     }
 }
 
 impl<Credential> AuthenticationProtocol for BasicAuthentication<Credential>
 where
-    Credential: AuthenticationCredentialUsernamePassword,
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedUsernamePassword,
 {
-    type Builder = http::request::Builder;
     type Request = hyper::Request<hyper::Body>;
     type Response = hyper::Response<hyper::Body>;
     type Error = hyper::Error;
 
-    fn configure(&self, builder: Self::Builder) -> Self::Builder {
-        let value = ::http_auth::basic::encode_credentials(
-            self.credential.username(),
-            self.credential.password(),
-        );
-        builder.set_sensitive_header(hyper::header::AUTHORIZATION, &value)
+    fn step(&self) -> Result<Option<AuthenticationStep<Self::Request>>, AuthenticError> {
+        match self.credential.auth_step() {
+            Ok(duration) if duration.is_zero() => Ok(None),
+            Ok(duration) => Ok(Some(AuthenticationStep::WaitFor(duration))),
+            Err(err) => Err(err),
+        }
+    }
+}
+
+impl<Credential> AuthenticationProtocolConfigure<http::request::Builder>
+    for BasicAuthentication<Credential>
+where
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedUsernamePassword,
+{
+    fn configure(
+        &self,
+        builder: http::request::Builder,
+    ) -> Result<http::request::Builder, AuthenticError> {
+        let fetched = self.credential.fetch()?;
+        let value = ::http_auth::basic::encode_credentials(fetched.username(), fetched.password());
+        let mut header_value = HeaderValue::try_from(value)?;
+        header_value.set_sensitive(true);
+        Ok(builder.header(hyper::header::AUTHORIZATION, header_value))
     }
 }
 
@@ -152,23 +212,23 @@ pub enum HttpAuthentication<Credential> {
 }
 
 impl<Credential> HttpAuthentication<Credential> {
-    pub fn new(credential: &Arc<HttpRealmCredentials<Credential>>) -> Self {
-        Self::Initial(credential.clone())
+    pub fn new(credential: Arc<HttpRealmCredentials<Credential>>) -> Self {
+        Self::Initial(credential)
     }
 }
 
 impl<Credential> AuthenticationProtocol for HttpAuthentication<Credential>
 where
-    Credential: AuthenticationCredentialUsernamePassword + 'static,
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedUsernamePassword,
 {
-    type Builder = http::request::Builder;
     type Request = hyper::Request<hyper::Body>;
     type Response = hyper::Response<hyper::Body>;
     type Error = hyper::Error;
 
-    fn step(&mut self) -> Option<AuthenticationStep<Self::Request>> {
+    fn step(&self) -> Result<Option<AuthenticationStep<Self::Request>>, AuthenticError> {
         match self {
-            Self::Initial(_) => None,
+            Self::Initial(_) => Ok(None),
             Self::Basic(basic) => basic.step(),
         }
     }
@@ -180,13 +240,6 @@ where
         }
     }
 
-    fn configure(&self, builder: Self::Builder) -> Self::Builder {
-        match self {
-            Self::Initial(_) => builder,
-            Self::Basic(basic) => basic.configure(builder),
-        }
-    }
-
     fn has_completed(&mut self, response: &Self::Response) -> Result<bool, AuthenticError> {
         match self {
             Self::Initial(realm_credentials) => {
@@ -200,9 +253,11 @@ where
                     match pw_client {
                         http_auth::PasswordClient::Basic(client) => {
                             let realm = client.realm();
-                            match realm_credentials.get_credential(realm) {
+                            let fetched = realm_credentials.fetch()?;
+                            match fetched.credential(realm) {
                                 Some(credential) => {
-                                    *self = Self::Basic(BasicAuthentication::new(credential));
+                                    *self =
+                                        Self::Basic(BasicAuthentication::new(credential.clone()));
                                     Ok(false)
                                 }
                                 None => Err(AuthenticError::UnknownRealm(realm.to_owned())),
@@ -219,3 +274,20 @@ where
         }
     }
 }
+
+impl<Credential> AuthenticationProtocolConfigure<http::request::Builder>
+    for HttpAuthentication<Credential>
+where
+    Credential: AuthenticationCredential,
+    <Credential as AuthenticationCredential>::Fetch: FetchedUsernamePassword,
+{
+    fn configure(
+        &self,
+        builder: http::request::Builder,
+    ) -> Result<http::request::Builder, AuthenticError> {
+        match self {
+            Self::Initial(_) => Ok(builder),
+            Self::Basic(basic) => basic.configure(builder),
+        }
+    }
+}
